/*
 *  yuvtorgb.c
 *
 * Convert YUV422 captured by NetWinder vidcap to RGB24.
 *
 *  Copyright (C) 1995-2000 Russell King.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <stdlib.h>

#include "yuvtorgb.h"

static int *xlat;

#define xlatu_b (xlat)
#define xlatu_g (xlat+256)
#define xlatv_g (xlat+512)
#define xlatv_r (xlat+768)

void yuv422_to_rgb_init(void)
{
	int i;

	xlat = malloc(256 * 4 * sizeof(int));

	for (i = 0; i < 256; i++) {
		int j;
		if (i < 16)
			j = 16 - 128;
		else if (i > 240)
			j = 240 - 128;
		else
			j = i - 128;

		xlatu_b[i] = (j * 219) / 126;
		xlatu_g[i] = (j *  74) / 219;
		xlatv_g[i] = (j * 153) / 219;
		xlatv_r[i] = (j * 219) / 160;
	}
}

static inline unsigned int xlaty(unsigned int y)
{
	if (y < 16)
		y = 16;
	if (y > 240)
		y = 240;
	return y;
}

static inline int limit(unsigned int y, int c)
{
	int v;

	__asm__ __volatile__("" : : : "memory");
	v = y + c;
	if (v < 0)
		v = 0;
	if (v > 255)
		v = 255;
	__asm__ __volatile__("" : : : "memory");
	return v;
}

void yuv422_to_rgb24(unsigned char *rgbbuf, unsigned char *yuvbuf, int sz)
{
	const unsigned int *src = (unsigned int *)yuvbuf;
	unsigned short *dst = (unsigned short *)rgbbuf;
	sz >>= 2;

	do {
		unsigned int val, xly0, xly1;
		int r, g, b;

		val = *src++;

		r  = xlatv_r[val >> 24];
		g  = xlatv_g[val >> 24];

		g += xlatu_g[(val >> 8) & 255];
		b  = xlatu_b[(val >> 8) & 255];

		xly0 = xlaty(val & 255);

		*dst++ = limit(xly0, r)  | limit(xly0, -g) << 8;

		xly1 = xlaty((val >> 16) & 255);

		*dst++ = limit(xly0, b)  | limit(xly1, r) << 8;
		*dst++ = limit(xly1, -g) | limit(xly1, b) << 8;
	} while (sz--);
}
